/*! \file halfmac_client_sw.c
 \brief Reception Half of a CSMA MAC protocol

 @version 14
 @author Chris Hunter

 The goal of this MAC is to act as a receiver to a CSMA-enabled server.
 This server will continue to retransmit packets until this client
 acknowledges its reception.

 In this variant, ACK transmission is handled directly by software using
 the transmission mechanisms in the WARPMAC framework.

 Note to workshop users: any names in quotes
 are names that can be searched for in the
 WARP API.

*/


#include "xparameters.h"
#include "warpmac.h"
#include "warpphy.h"
#include "halfmac_client_sw.h"
#include "string.h"
#include "errno.h"
#include "stdlib.h"
#include "stdio.h"


//Define handy macros for CSMA MAC packet types
///Data packet with payload meant for Ethernet transmission
#define DATAPACKET 1
///Acknowledgement packet meant for halting retransmissions
#define ACKPACKET 0

///Buffer for holding a packet-to-xmit across multiple retransmissions
Macframe txMacframe;
///Buffer to hold received packet
Macframe rxMacframe;

unsigned short int myID;
unsigned int autoResp_matchCond;
unsigned int autoResp_action;
unsigned char pktBuf_tx_ACK;
unsigned char pktBuf_tx_DATA;
unsigned char pktBuf_rx;


///@brief Callback for the reception of Ethernet packets
///
///This function is called by the ethernet MAC drivers
///when a packet is available to send. This function fills
///the Macframe transmit buffer with the packet and sends
///it over the OFDM link
///@param length Length, in bytes, of received Ethernet frame
///@param payload address of first byte in Ethernet payload.
void dataFromNetworkLayer_callback(Xuint32 length, char* payload)
{

	//Note: This code is virtually identical to the code used previously in the noMac workshop exercise.
	//It does not fully implement the halfmac_server's transmission states (e.g. no retransmission, no carrier sensing,
	//no binary exponential backoff).

	//Set the length field in the header
	txMacframe.header.length = length;
	//Set the type to be a data packet
	txMacframe.header.pktType = DATAPACKET;
	//Copy in the packet's destination MAC address
	txMacframe.header.destAddr = (unsigned short int)(NODEID_TO_ADDR(0));
	//Set the modulation scheme for the packet's full-rate symbols
	txMacframe.header.fullRate = HDR_FULLRATE_QPSK;
	//Set the payload coding rate
	txMacframe.header.codeRate = CODE_RATE_NONE;
	//Copy the header over to packet buffer 1
	warpmac_prepPhyForXmit(&txMacframe, pktBuf_tx_DATA);
	//Send packet buffer 1
	warpmac_startPhyXmit(1);
	//Wait for it to finish and enable the receiver
	warpmac_finishPhyXmit();
}

///@brief Callback for the reception of bad wireless headers
///
///@param packet Pointer to received Macframe
void phyRx_badHeader_callback()
{
	warpmac_incrementLEDLow();
}

///@brief Callback for the reception of good wireless headers
///
///This function then polls the PHY to determine if the entire packet passes checksum
///thereby triggering the transmission of the received data over Ethernet.
///@param packet Pointer to received Macframe
void phyRx_goodHeader_callback(Macframe* packet)
{

//WORKSHOP PSEUDOCODE:
//1) Instantiate a char variable to keep track of the receiver's state. Default this state variable to "PHYRXSTATUS_INCOMPLETE"
//2) Instantiate a new Macframe that will be the acknowledgment packet you send when and if the payload of this packet is deemed good
//3) Check the 'destAddr' element in the 'header' struct of the 'packet' Macframe. Only proceed if this value matches 'myID'
//			Note: myID is a global that takes a value down in the main function of this program. For convenience, we have tied
//			a node's ID to the value of the dip-switch below the hex display on the WARP FPGA board.
//4) Check the 'pktType' field of the header.
//If 'DATAPACKET'
//		5) Fill in the Macframe you created at the top of this function. You must fill in the following fields:
//			- the 'length' should be 0 (i.e. there is no payload present in an ACK packet)
//			- the 'pktType' should be 'ACKPACKET' in order to differentiate this packet from data frames
//			- the 'fullRate' does not really matter since there is no payload in this packet, but just set it to 'HDR_FULLRATE_QPSK'
//		6) Copy over the 6-byte 'srcAddr' field of the received packet into the 6-byte 'destAddr' field of the ack packet using a standard memcpy call
//		7) Copy the ACK into the 'pktBuf_tx_ACK' PHY buffer using "warpmac_prepPhyForXmit"
//		8) Poll the state of the state of the receiver using "warpmac_finishPhyRecv." Block until the state turns to either "PHYRXSTATUS_GOOD" or "PHYRXSTATUS_BAD"
//		If "GOODPACKET"
//			9a) Send the ACK using 'warpmac_startPhyXmit'
//			10a) Animate the top two LEDs to visualize this behavior using the "warpmac_incrementLEDHigh" function
//			11a) Copy the received "Macframe" to the Ethernet MAC (Emac) using "warpmac_prepPktToNetwork"
//					Note: The first argument of this function is the beginning of the packet that you want sent over the wire.
//					  	  This does NOT include all of the extra wireless MAC header information of the packet. The first byte
//					  	   of the payload is located at (void *)warpphy_getBuffAddr(pktBuf_rx)+NUM_HEADER_BYTES,
//					  	  where pktBuf_rx is an already defined global in this file (noMac.c) that specifies the location of
//					  	  the Macframe in the PHY.
//			12a) Wait for the ACK to finish sending with "warpmac_finishPhyXmit"
//			13a) Start the Emac using "warpmac_startPktToNetwork"
//				Note: The only argument to this function is the length (in bytes) of the packet to be sent. This length is stored in the
//				the 'length' field of the 'header' struct belonging to the 'packet' Macframe (i.e. packet->header.length).
//		If "BADPACKET"
//			9b) Animate the bottom two LEDs to visualize this behavior using the "warpmac_incrementLEDLow" function


/**********************DELETE FOR WORKSHOP***************************/
	unsigned char state = PHYRXSTATUS_INCOMPLETE;
	Macframe ackPacket;

	//If the packet is addressed to this node
	if( packet->header.destAddr == (NODEID_TO_ADDR(myID)) )
	{
		switch(packet->header.pktType){
			//If received packet is data
			case DATAPACKET:

				ackPacket.header.length = 0;
				ackPacket.header.pktType = ACKPACKET;
				ackPacket.header.fullRate = HDR_FULLRATE_QPSK;
				ackPacket.header.codeRate = CODE_RATE_NONE;
				memcpy(ackPacket.header.destAddr,packet->header.srcAddr,6);

				//Copy the header over to packet pktBuf_tx_ACK
				warpmac_prepPhyForXmit(&ackPacket,pktBuf_tx_ACK);

				//Blocks until the PHY declares the payload good or bad
				state = warpmac_finishPhyRecv();

				if(state & PHYRXSTATUS_GOOD){
					warpmac_startPhyXmit(pktBuf_tx_ACK);

					//Toggle the top LEDs
					warpmac_incrementLEDHigh();

					//Starts the DMA transfer of the payload into the EMAC
					warpmac_prepPktToNetwork((void *)warpphy_getBuffAddr(pktBuf_rx)+NUM_HEADER_BYTES, (packet->header.length));

					//Blocks until the PHY is finished sending and enables the receiver
					warpmac_finishPhyXmit();

					//Waits until the DMA transfer is complete, then starts the EMAC
					warpmac_startPktToNetwork((packet->header.length));


				}

				if(state & PHYRXSTATUS_BAD){
					warpmac_incrementLEDLow();
				}

				break;

		}
	}
/**********************DELETE FOR WORKSHOP***************************/
}

///@brief Main function
///
///This function configures MAC parameters, enables the underlying frameworks, and then loops forever.
int main()
{
	xil_printf("HALFMAC Client: Software-driven Acknowledgments\r\n");


	//Assign the packet buffers in the PHY
	//The auto responder can't transmit from buffer 0, so we use it for Rx packets
	//The other assignments (DATA/ACK) are arbitrary; any buffer in [1,30] will work
	pktBuf_rx = 0;
	pktBuf_tx_DATA = 1;
	pktBuf_tx_ACK = 2;

	//Initialize the framework
	// This function sets safe defaults for many parameters in the MAC/PHY frameworks
	// Many of these can be changed with other warpmac_ and warpphy_ calls
	//  or by customizing the warpmac.c/warpphy.c source
	warpmac_init();

	//Read Dip Switch value from FPGA board.
	//This value will be used as an index into the routing table for other nodes
	myID = (unsigned short int)warpmac_getMyId();
	warpmac_rightHex(myID);

	//Set the PHY for SISO using just the radio in slot 2
	warpphy_setAntennaMode(TX_ANTMODE_SISO_ANTA, RX_ANTMODE_SISO_ANTA);

	//Rx buffer is where the EMAC will DMA Wireless payloads from
	warpmac_setRxBuffers(&rxMacframe, pktBuf_rx);

	//Tx buffer is where the EMAC will DMA Ethernet payloads to
	warpmac_setPHYTxBuffer(pktBuf_tx_DATA);

	//Connect the various user-level callbacks
	warpmac_setCallback(EVENT_DATAFROMNETWORK, (void *)dataFromNetworkLayer_callback);
	warpmac_setCallback(EVENT_PHYGOODHEADER, (void *)phyRx_goodHeader_callback);
	warpmac_setCallback(EVENT_PHYBADHEADER, (void *)phyRx_badHeader_callback);

	//Set the default center frequency
	warpphy_setChannel(GHZ_2, 3);

	//Enable Ethernet
	warpmac_enableDataFromNetwork();

	while(1)
	{
		//Poll the timer, PHY and user I/O forever; actual processing will happen via callbacks above
		warpmac_pollPeripherals();
	}

	return;
}
